# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import os

def expect_error(func):
    try:
        func()
    except Exception as e:
        return
    raise Exception("Expect an error, but nothing catched.")

class TestCore(unittest.TestCase):
    def test_number_of_hold_vars(self):
        assert jt.random([1,2,3]).peek() == "float32[1,2,3,]"
        assert jt.core.number_of_hold_vars() == 0
        x = jt.random([1,2,3])
        assert jt.core.number_of_hold_vars() == 1
        del x
        assert jt.core.number_of_hold_vars() == 0

    def test_fetch_sync(self):
        dtypes = ["float32", "float64"]
        for dtype in dtypes:
            x = jt.random([1,2,3], dtype)
            res = x.data
            assert res.dtype == dtype and res.shape == (1,2,3)

    def test_set_seed(self):
        a = jt.random([1,2,3]).data
        b = jt.random([1,2,3]).data
        assert str(a) != str(b)
        jt.set_seed(1)
        a = jt.random([1,2,3]).data
        jt.set_seed(1)
        b = jt.random([1,2,3]).data
        assert str(a) == str(b)
        
    def test_array_op(self):
        data = [
            np.array([1,2,3]),
            np.int32([1,2,3]),
            np.int64([1,2,3]),
            np.float32([1,2,3]),
            np.float64([1,2,3]),
        ]
        for a in data:
            assert sum(jt.array(a).data) == 6
        assert np.all(jt.array(np.int32([1,2,3])[::-1]).data == [3,2,1])
        assert jt.array(1).data.shape == (1,)
        
    def test_matmul_op(self):
        a = np.array([[1, 0], [0, 1]]).astype("float32")
        b = np.array([[4, 1], [2, 2]]).astype("float32")
        c = np.matmul(a, b)
        jtc = jt.matmul(jt.array(a), jt.array(b)).data
        assert np.allclose(jtc, c)

        a = np.random.random((128,3,10,20))
        b = np.random.random((20,30))
        c = np.matmul(a, b)
        jtc = jt.matmul(jt.array(a), jt.array(b)).data
        assert np.allclose(jtc, c)

        a = np.random.random((128,3,10,20))
        b = np.random.random((128,3,20,30))
        c = np.matmul(a, b)
        jtc = jt.matmul(jt.array(a), jt.array(b)).data
        assert np.allclose(jtc, c), np.abs(jtc-c).max()
        
    def test_var_holder(self):
        jt.clean()
        expect_error(lambda: jt.matmul(1,1))
        expect_error(lambda: jt.matmul([1],[1]))
        expect_error(lambda: jt.matmul([[1]],[1]))
        self.assertEqual(jt.number_of_lived_vars(), 0)
        a = jt.matmul(jt.float32([[3]]), jt.float32([[4]])).data
        assert a.shape == (1,1) and a[0,0] == 12
        a = np.array([[1, 0], [0, 1]]).astype("float32")
        b = np.array([[4, 1], [2, 2]]).astype("float32")
        c = np.matmul(a, b)
        jtc = jt.matmul(jt.array(a), jt.array(b)).data
        assert np.all(jtc == c)
        
    def test_save_load_sub_module(self):
        class Net(jt.Module):
            def __init__(self):
                self.conv1 = jt.nn.Conv(3,3,3)
        net = Net()
        assert list(net.state_dict().keys()) == ['conv1.weight', 'conv1.bias']
        assert list(net.conv1.state_dict().keys()) == ['weight', 'bias']
        pkl_name = os.path.join(jt.flags.cache_path, "sub.pkl")
        net.conv1.save(pkl_name)
        net.conv1.load(pkl_name)

    def test_module(self):
        a = jt.Module()
        a.__setattr__("x", 1)
        assert a.__getattr__("x") == 1
        a.y = 2
        assert a.y == 2

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import os
import re
import sys

class TestJtune(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        n, m, k = 2, 6, 16
        a = jt.random((n, m, 1))
        b = jt.random((1, m, k))
        jt.fetch_sync([a,b])
        with jt.profile_scope(
            compile_options = {"jtune":1}
        ) as rep:
            c = (a*b).sum(1)
            c.sync()
        assert len(rep) == 2
        self.fname = rep[1][1]
        self.jtune_path = os.path.join(jt.flags.jittor_path, "utils/jtune.py")

    def run_cmd(self, cmd):
        cmd = f"warmup=0 rerun=0 {sys.executable} {self.jtune_path} {self.fname} {cmd}"
        return jt.compiler.run_cmd(cmd)

    def test_run_so(self):
        res = self.run_cmd("run_so").splitlines()
        assert res[0]=="Enter fake_main entry.", res
        assert res[1]=="     Count TotalTime   AvgTime   MinTime   MaxTime     Input    Output   Compute", res
        nums = res[2].split()
        assert nums[0]=="1", nums

    def test_cc_to_so(self):
        self.run_cmd("cc_to_so")

    def test_cc_to_s(self):
        self.run_cmd("cc_to_s")
        sname = self.fname[:-2] + 's'
        with open(sname) as f:
            src = f.read()
        fma_ins = re.findall("fma.*", src)
        assert len(fma_ins)>=4, f"fma instructions should be used for matmul. {fma_ins}"
        self.run_cmd("s_to_so")
        

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: 
#     Guowei Yang <471184555@qq.com>
#     Meng-Hao Guo <guomenghao1997@gmail.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
from jittor import nn, Module
from jittor.models import inception
import numpy as np
import sys, os
import random
import math
import unittest
from jittor.test.test_reorder_tuner import simple_parser
from jittor.test.test_log import find_log_with_re
from jittor.dataset.mnist import MNIST
import jittor.transform as trans
import time

skip_this_test = False

class MnistNet(Module):
    def __init__(self):
        self.model = inception.inception_v3()
        self.layer = nn.Linear(1000,10)
    def execute(self, x):
        x = self.model(x)
        x = self.layer(x)
        return x

@unittest.skipIf(skip_this_test, "skip_this_test")
class TestInception(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # hyper-parameters
        self.batch_size = 32
        self.weight_decay = 0.0001
        self.momentum = 0.9
        self.learning_rate = 0.1
        # mnist dataset
        self.train_loader = MNIST(train=True, transform=trans.Resize(300)) \
            .set_attrs(batch_size=self.batch_size, shuffle=True)
        self.train_loader.num_workers = 4
        self.train_loader.total_len = self.batch_size * 300

    # setup random seed
    def setup_seed(self, seed):
        np.random.seed(seed)
        random.seed(seed)
        jt.seed(seed)

    @unittest.skipIf(not jt.has_cuda, "Cuda not found")
    @jt.flag_scope(use_cuda=1, use_stat_allocator=1)
    def test_inception(self):
        self.setup_seed(1)
        loss_list=[]
        acc_list=[]
        mnist_net = MnistNet()
        global prev
        prev = time.time()
        SGD = nn.SGD(mnist_net.parameters(), self.learning_rate, self.momentum, self.weight_decay)

        for batch_idx, (data, target) in enumerate(self.train_loader):

            # train step
            with jt.log_capture_scope(
                log_silent=1,
                log_v=1, log_vprefix="op.cc=100,exe=10",
            ) as logs:
                # breakpoint()
                output = mnist_net(data)
                loss = nn.cross_entropy_loss(output, target)
                SGD.step(loss)
                def callback(batch_idx, loss, output, target):
                    # print train info
                    global prev
                    pred = np.argmax(output, axis=1)
                    acc = np.mean(target==pred)
                    loss_list.append(loss[0])
                    acc_list.append(acc)
                    print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tAcc: {:.6f} \tTime:{:.3f}'
                        .format(0, batch_idx, 300,1. * batch_idx / 6.0, loss[0], acc, time.time()-prev))
                    # prev = time.time()
                jt.fetch(batch_idx, loss, output, target, callback)
            
            log_conv = find_log_with_re(logs, 
                "Jit op key (not )?found: ((mkl)|(cudnn))_conv.*")
            log_matmul = find_log_with_re(logs, 
                "Jit op key (not )?found: ((mkl)|(cublas))_matmul.*")
            if batch_idx > 2:
                assert len(log_conv)==283 and len(log_matmul)==6, (len(log_conv), len(log_matmul))

            mem_used = jt.flags.stat_allocator_total_alloc_byte \
                -jt.flags.stat_allocator_total_free_byte
            # assert mem_used < 4e9, mem_used
            # TODO: why bigger?
            assert mem_used < 15.6e9, mem_used
            # example log:
            # Train Epoch: 0 [0/100 (0%)]     Loss: 2.352903  Acc: 0.110000
            # Train Epoch: 0 [1/100 (1%)]     Loss: 2.840830  Acc: 0.080000
            # Train Epoch: 0 [2/100 (2%)]     Loss: 3.473594  Acc: 0.100000
            # Train Epoch: 0 [3/100 (3%)]     Loss: 3.131615  Acc: 0.200000
            # Train Epoch: 0 [4/100 (4%)]     Loss: 2.524094  Acc: 0.230000
            # Train Epoch: 0 [5/100 (5%)]     Loss: 7.780025  Acc: 0.080000
            # Train Epoch: 0 [6/100 (6%)]     Loss: 3.890721  Acc: 0.160000
            # Train Epoch: 0 [7/100 (7%)]     Loss: 6.370137  Acc: 0.140000
            # Train Epoch: 0 [8/100 (8%)]     Loss: 11.390827 Acc: 0.150000
            # Train Epoch: 0 [9/100 (9%)]     Loss: 21.598564 Acc: 0.080000
            # Train Epoch: 0 [10/100 (10%)]   Loss: 23.369165 Acc: 0.130000
            # Train Epoch: 0 [20/100 (20%)]   Loss: 4.804510  Acc: 0.100000
            # Train Epoch: 0 [30/100 (30%)]   Loss: 3.393924  Acc: 0.110000
            # Train Epoch: 0 [40/100 (40%)]   Loss: 2.286762  Acc: 0.130000
            # Train Epoch: 0 [50/100 (50%)]   Loss: 2.055014  Acc: 0.290000

            assert jt.core.number_of_lived_vars() < 50000, jt.core.number_of_lived_vars()

        jt.sync_all(True)
        assert np.mean(loss_list[-20:])<1
        assert np.mean(acc_list[-20:])>0.5
        
if __name__ == "__main__":
    unittest.main()



import os
import sys
import subprocess as sp

def_path = sys.argv[-1]

# print(sys.argv)
dumpbin_path = os.environ.get("dumpbin_path", "dumpbin")
export_all = os.environ.get("EXPORT_ALL", "0")=="1"

syms = {}

for obj in sys.argv[1:-2]:
    cmd = f'"{dumpbin_path}" -SYMBOLS "{obj}"'
    ret = sp.getoutput(cmd)
    # print(ret)
    for l in ret.splitlines():
        if '|' in l:
            if "UNDEF" in l: continue
            if "External" not in l: continue
            sym = l.split('|')[1].strip().split()[0]
            if sym[0] in '@.': continue
            if sym.startswith("??$get_from_env"): syms[sym] = 1
            # if sym.startswith("??"): continue
            if sym.startswith("my"): syms[sym] = 1
            # for cutt
            if "custom_cuda" in sym: syms[sym] = 1
            if "cutt" in sym: syms[sym] = 1
            if "_cudaGetErrorEnum" in sym: syms[sym] = 1
            if export_all: syms[sym] = 1
            if "jittor" not in sym: continue
            syms[sym] = 1
    # print(ret)
libname = os.path.basename(def_path).rsplit(".", 1)[0]
src = f"LIBRARY {libname}\nEXPORTS\n"
for k in syms:
    src += f"    {k}\n"
# print(src)

with open(def_path, "w") as f:
    f.write(src)



# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved.
# Maintainers:
#     Zheng-Ning Liu <lzhengning@com>
#
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

""" This file implements generation of stub files for Jittor C extensions.

In detail, autocompletion of the following functions are supported.
- functions in __init__.py
- functions in jittor.core.ops
- attributes of jittor.flags
- methods of jittor.Var

Prerequisite:
- mypy for automatic stub generation

Usage: python3 -m jittor.utils.gen_pyi

"""

import os
import re
import shutil
import jittor

def add_indent(s: str, n=1):
    for _ in range(n):
        s = '\t' + s.replace('\n', '\n\t', s.count('\n')-1)
    return s

def ctype_to_python(type_str):
    if type_str == "bool":
        return "bool"
    if type_str in ["int", "uint", "int64", "uint64", "size_t"]:
        return "int"
    if type_str in ["float32", "float64"]:
        return "float"
    if type_str in ["string", "string&&", "NanoString", "char*", "const char*"]:
        return "str"
    if type_str in ["vector<int>"]:
        return "List[int]"
    if type_str in ["vector<string>&&", "vector<NanoString>&&"]:
        return "List[str]"
    if type_str == "VarHolder*":
        return "Var"
    if type_str in ["vector<VarHolder*>", "vector<VarHolder*>&&"]:
        return "List[Var]"
    if type_str == "NanoVector":
        return "Tuple[int]"
    if type_str == "vector<NanoVector>&&":
        return "List[Tuple[int]]"
    if type_str in ["FetchFunc", "FetchFunc&&", "NumpyFunc&&"]:
        return "Callable"
    if type_str == "vector<NumpyFunc>&&":
        return "List[Callable]"
    if type_str == "PyObject*":
        return "float | int | numpy.ndarray | Var"
    if type_str == "VarSlices&&":
        return "slice"
    if type_str in ["ArrayArgs", "ArrayArgs&&", "DataView"]:
        return "numpy.ndarray"
    if type_str == 'ItemData':
        return "float | int | bool"
    if type_str == "void":
        return ""
    print(f"[warning] Unknown ctype: {type_str}, do not write type hinting")
    return ""

def cval_to_python(val_str: str):
    if val_str == "false":
        return "False"
    if val_str == "true":
        return "True"
    if val_str.startswith("ns_"):
        return f'"{val_str[3:]}"'
    if val_str == "NanoVector()":
        return "()"
    return val_str


def run_stubgen(jittor_path, cache_path):

    # for __init__.py functions
    stubpath = os.path.join(cache_path, 'stubs')
    stubfile = os.path.join(stubpath, "jittor", "__init__.pyi")
    os.system(f"stubgen -m jittor -o {stubpath} -q")
    with open(stubfile) as f:
        mypy_content = f.read()

    f = open(stubfile, "w")
    # Remove the follow type redirection
    unused_content = ["ori_int = int\n",
                      "ori_float = float\n",
                      "ori_bool = bool\n",
                      "int = int32\n",
                      "float = float32\n",
                      "double = float64\n",
                      "\nflags: Any\n"]
    for unused in unused_content:
        mypy_content = mypy_content.replace(unused, "")
    f.write(mypy_content)

    shutil.move(stubfile, os.path.join(jittor_path, "__init__.pyi"))
    shutil.rmtree(stubpath)
    shutil.rmtree(os.path.expanduser(".mypy_cache"))

def gen_ops_stub(jittor_path):
    f = open(os.path.join(jittor_path, "__init__.pyi"), "a")
    f.write("from typing import List, Tuple, Callable, overload\n")
    f.write("import numpy\n")

    var_hint = "class Var:\n\t'''Variable that stores multi-dimensional data.'''\n"
    var_methods = set()

    def decl_to_param_hints(decl):
        param_decl = re.findall(r".+ [a-zA-Z_0-9]+\((.*)\)", decl)[0]
        if not param_decl.strip():
            return []
        param_hints = []
        for param_str in param_decl.split(','):
            if "=" in param_str:
                template = r"\s*(.+)\s+([a-zA-Z_0-9]+)\s*=\s*(.+)"
                param_type, param_name, param_val = re.findall(template, param_str)[0]
                param_type = ctype_to_python(param_type)
                param_val = cval_to_python(param_val)
            else:
                param_type, param_name = param_str.strip().rsplit(' ', maxsplit=1)
                param_type = ctype_to_python(param_type)
                param_val = ""

            hint = param_name
            if param_type:
                hint += ": " + param_type
            if param_val:
                hint += "=" + param_val
            param_hints.append(hint)
        return param_hints

    def generate_var_hint(decorators, return_type, param_hints, docstring):
        hint = add_indent(decorators) if decorators else ""
        hint += f"\tdef {func_name}("
        hint += ", ".join(['self'] + param_hints) + ")"
        hint += f"-> {return_type}" if return_type else ""
        hint += ":"
        if docstring:
            hint += add_indent(f"\n'''{docstring}'''\n", 2) + "\t\t...\n"
        else:
            hint += f" ...\n"
        return hint

    for func_name, func in jittor.ops.__dict__.items():
        if func_name.startswith("__"):
            continue
        # Exclude a function that overrides the builtin bool:
        #       def bool(x: Var) -> Var: ...
        # It will confuse the IDE. So we ignore this function in pyi.
        if func_name == "bool":
            continue

        docstring = func.__doc__[:func.__doc__.find("Declaration:")]
        docstring = docstring.replace("'''", '"""').strip()
        declarations = re.findall(r"Declaration:\n(.+)\n", func.__doc__)

        for decl in declarations:
            decorators = "@overload\n" if len(declarations) > 1 else ""
            return_type = ctype_to_python(decl.split(' ', maxsplit=1)[0])
            param_hints = decl_to_param_hints(decl)

            func_text = decorators
            func_text += f"def {func_name}"
            func_text += "(" + ", ".join(param_hints) + ")"
            func_text += f"-> {return_type}" if return_type else ""
            func_text += ":\n"
            if docstring:
                func_text += add_indent(f"'''{docstring}'''\n") + "\t...\n"
            else:
                func_text += f" ...\n"

            f.write(func_text)

            if not "Var" in param_hints[0]:
                continue
            var_methods.add(func_name)
            var_hint += generate_var_hint(decorators, return_type, param_hints[1:], docstring)

    for func_name, func in jittor.Var.__dict__.items():
        if func_name.startswith("__") or func_name in var_methods:
            continue
        if func_name in ["int", "float", "double", "bool", "long"]:
            continue
        if func.__doc__ is None:
            continue
        docstring = func.__doc__[:func.__doc__.find("Declaration:")]
        docstring = docstring.replace("'''", '"""').strip()
        declarations = re.findall(r"Declaration:\n(.+)\n", func.__doc__)

        for decl in declarations:
            decl = decl.replace("inline ", "")
            decorators = "@overload\n" if len(declarations) > 1 else ""
            return_type = re.findall(r"(.+) [a-zA-Z_0-9]+\(.*\)", decl)[0].split()[-1]
            return_type = ctype_to_python(return_type)
            param_hints = decl_to_param_hints(decl)

            var_hint += generate_var_hint(decorators, return_type, param_hints, docstring)

    f.write(var_hint)
    f.close()

def gen_flags_stub(jittor_path):
    f = open(os.path.join(jittor_path, "__init__.pyi"), "a")
    f.write("class Flags:\n")
    f.write("\t'''A set of flags to configure jittor running behaviors'''\n")

    for attr_name, attr in jittor.Flags.__dict__.items():
        if attr_name.startswith("__"):
            continue
        docstring = attr.__doc__
        docstring = attr.__doc__[:attr.__doc__.find("Declaration:")]
        docbody = re.findall("\(type.+default.+\):(.+)", docstring)[0].strip()
        docbody += "." if not docbody.endswith('.') else ""
        attr_type, attr_val = re.findall(r"\(type:(.+), default:(.+)\)", docstring)[0]
        attr_type = ctype_to_python(attr_type)
        attr_type = attr_type if attr_type else "Any"
        f.write(f"\t{attr_name}: {attr_type}\n")
        f.write(f"\t'''{docbody} Default: {attr_val}'''\n")

    f.write("flags: Flags\n")
    f.write("'''Jittor running time flags instance'''\n")
    f.close()

def get_pyi(jittor_path=None, cache_path=None):
    if jittor_path is None:
        jittor_path = jittor.flags.jittor_path
    if cache_path is None:
        import jittor_utils
        cache_path = jittor_utils.cache_path

    run_stubgen(jittor_path, cache_path)
    gen_ops_stub(jittor_path)
    gen_flags_stub(jittor_path)

    print(f"Generated stubfile: {os.path.join(jittor_path, '__init__.pyi')}")


if __name__ == "__main__":
    get_pyi()


# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_core import expect_error
from .test_grad import ngrad
from .test_cuda import test_cuda

def all_eq(x, y):
    if len(x.shape) == 0: x = np.array([x])
    if len(y.shape) == 0: y = np.array([y])
    convert = lambda x: x.astype("uint8") if x.dtype=="bool" else x
    x = convert(x)
    y = convert(y)
    if str(x.dtype).startswith("float"):
        return str(y.dtype).startswith("float") and x.shape == y.shape and (x==y).all()
    return x.dtype == y.dtype and x.shape == y.shape and (x==y).all()

def check(op, *args):
    x = eval(f"np.{op}(*args)")
    y = eval(f"jt.{op}(*args).data")
    assert all_eq(x, y), f"{x}\n{y}"

class TestBinaryOp(unittest.TestCase):
    def test_binary_op(self):
        assert np.all(jt.binary(1,2,'maximum').data == 2)
        assert np.all(jt.binary([[1,2]],[[3,4]],'add').data == [[4,6]])
        assert np.all(jt.less(1,2).data)
        assert jt.less(1,2).data.dtype == "bool"
        x = (jt.array(1) << jt.array(3)).data
        assert (x == 8).all()
        x = (jt.array(2) ** jt.array(3)).data
        assert (x == 8).all()
        a = np.array([1,2,3])
        b = np.array([7,10,13])
        check("logical_and", a, b)
        check("logical_or", a, b)
        check("logical_xor", a, b)
        check("bitwise_and", a, b)
        check("bitwise_or", a, b)
        check("bitwise_xor", a, b)
        
    def test_i(self):
        def check(op, a, b):
            if jt.flags.use_cuda and op == "@":
                return
            if op=="@":
                a = np.float32(a)
                b = np.float32(b)
            ja = jt.array(a)
            jb = jt.array(b)
            exec(f"ja {op}= jb")
            ja = ja.fetch_sync()
            
            if op == "@":
                # numpy do not support @=
                a = np.array(a) @ np.array(b)
            else:
                a = eval(f"a {op} b")
                a = np.float32(a)
            ja = np.float32(ja)
            
            assert all_eq(ja, a), (ja,a)
        check("+", 5, 2)
        check("-", 5, 2)
        check("*", 5, 2)
        check("/", 5, 2)
        check("//", 5, 2)
        check("@", [[5]], [[2]])
        check("%", 5, 2)
        check("**", 5, 2)
        check("<<", 5, 2)
        check(">>", 5, 2)
        check("&", 5, 2)
        check("^", 5, 2)
        check("|", 5, 2)
        
    def test_r(self):
        def check(op, a, b):
            a = np.array(a)
            b = np.array(b)
            if jt.flags.use_cuda and op == "@":
                return
            jb = jt.array(b)
            jc = eval(f"a {op} jb").data

            
            if op == "@":
                # numpy do not support @=
                a = np.array(a) @ np.array(b)
            else:
                a = eval(f"a {op} b")
                a = np.array(a)
            
            assert all_eq(jc, a), f"\n{jc}\n{a}"
        check("+", 5, 2)
        check("-", 5, 2)
        check("*", 5, 2)
        check("/", 5, 2)
        check("//", 5, 2)
        # check("@", [[5]], [[2]])
        check("%", 5, 2)
        check("**", 5, 2)
        check("<<", 5, 2)
        check(">>", 5, 2)
        check("&", 5, 2)
        check("^", 5, 2)
        check("|", 5, 2)
        
    def test_grad(self):
        ops = ["+", "-", "*", "/", "**"]
        np.random.seed(3)
        a = np.random.rand(10)
        b = np.random.rand(10)
        c = np.random.rand(10)
        for op in ops:
            func = lambda x: eval(f"((x[0]{op}x[1])*x[2]).sum()")
            x, grads = ngrad(func, [a,b,c], 1e-8)
            ja = jt.array(a).name("ja")
            jb = jt.array(b).name("jb")
            jc = jt.array(c).name("jc")
            jx = eval(f"(ja{op}jb)*jc")
            jgrads = jt.grad(jx, [ja,jb,jc])
            for jd, nd in zip(jgrads, grads):
                assert (np.abs(jd.data-nd)<1e-4).all(), f"\n{jd.data}\n{nd}"

    def test_mod_float(self):
        a = jt.random((10,))
        b = jt.random((10,))
        c = a % b
        assert np.allclose(c.data, a.data % b.data)
        a = jt.random((10,), 'float64')
        b = jt.random((10,), 'float64')
        c = a % b
        assert np.allclose(c.data, a.data % b.data)
        a = jt.random((10,)) * 1000
        b = (jt.random((10,)) * 10).int() + 1
        c = a % b
        assert np.allclose(c.data, a.data % b.data), (c.data, a.data%b.data)

    def test_mod_grad(self):
        a = jt.random((100,))
        b = jt.random((100,))
        c = a % b
        da, db = jt.grad(c, [a, b])
        np.testing.assert_allclose(da.data, 1)
        np.testing.assert_allclose(db.data, -np.floor(a.data/b.data))

    def test_mod_negtive(self):
        a = jt.random((100,)) - 0.5
        b = jt.random((100,)) - 0.5
        c = a % b
        nc = a.data % b.data
        np.testing.assert_allclose(c.data, nc.data, atol=1e-5, rtol=1e-5)
    
    def test_pow(self):
        # win cuda 10.2 cannot pass
        a = jt.random((100,))
        b = a**3
        b.sync()



class TestBinaryOpCuda(TestBinaryOp, test_cuda(2)):
    pass

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2021 Jittor.
# All Rights Reserved. 
# Maintainers:
#     Dun Liang <randonlang@gmail.com>. 
#
# Contributors:
#     Xin Yao <yaox12@outlook.com>
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
from typing import Sequence
from PIL import Image, ImageOps, ImageEnhance, __version__ as PILLOW_VERSION
import numpy as np
import numbers
import math
from math import cos, sin, tan


def _is_pil_image(img):
    return isinstance(img, Image.Image)


def _get_image_size(img):
    if _is_pil_image(img):
        return img.size
    raise TypeError(f"Unexpected type {type(img)}")


def _get_image_num_channels(img):
    if _is_pil_image(img):
        return 1 if img.mode == 'L' else 3
    raise TypeError(f"Unexpected type {type(img)}")


def hflip(img):
    """
    Function for horizontally flipping the given image.

    Args::

        [in] img(PIL Image.Image): Input image.

    Example::
        
        img = Image.open(...)
        img_ = transform.hflip(img)
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    return img.transpose(Image.FLIP_LEFT_RIGHT)


def vflip(img):
    """
    Function for vertically flipping the given image.

    Args::

        [in] img(PIL Image.Image): Input image.

    Example::
        
        img = Image.open(...)
        img_ = transform.vflip(img)
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    return img.transpose(Image.FLIP_TOP_BOTTOM)


def adjust_brightness(img, brightness_factor):
    """
    Function for adjusting brightness of an RGB image.

    Args::

        [in] img (PIL Image.Image): Image to be adjusted.
        [in] brightness_factor (float):  How much to adjust the brightness.
             Can be any non negative number. 0 gives a black image, 1 gives the
             original image while 2 increases the brightness by a factor of 2.

    Returns::

        [out] PIL Image.Image: Brightness adjusted image.

    Example::
        
        img = Image.open(...)
        img_ = transform.adjust_brightness(img, 0.5)
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    enhancer = ImageEnhance.Brightness(img)
    img = enhancer.enhance(brightness_factor)
    return img


def adjust_contrast(img, contrast_factor):
    """
    Function for adjusting contrast of an image.

    Args::

        [in] img (PIL Image.Image): Image to be adjusted.
        [in] contrast_factor (float): How much to adjust the contrast.
             Can be any non negative number. 0 gives a solid gray image,
             1 gives the original image while 2 increases the contrast by a factor of 2.

    Returns::

        [out] PIL Image.Image: Contrast adjusted image.

    Example::
        
        img = Image.open(...)
        img_ = transform.adjust_contrast(img, 0.5)
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    enhancer = ImageEnhance.Contrast(img)
    img = enhancer.enhance(contrast_factor)
    return img


def adjust_saturation(img, saturation_factor):
    """
    Function for adjusting saturation of an image.

    Args::

        [in] img (PIL Image.Image): Image to be adjusted.
        [in] saturation_factor (float):  How much to adjust the saturation.
             0 will give a black and white image, 1 will give the original image
             while 2 will enhance the saturation by a factor of 2.

    Returns::

        [out] PIL Image.Image: Saturation adjusted image.

    Example::
        
        img = Image.open(...)
        img_ = transform.adjust_saturation(img, 0.5)
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    enhancer = ImageEnhance.Color(img)
    img = enhancer.enhance(saturation_factor)
    return img


def adjust_hue(img, hue_factor):
    """
    Function for adjusting hue of an image.

    The image hue is adjusted by converting the image to HSV and
    cyclically shifting the intensities in the hue channel (H).
    The image is then converted back to original image mode.

    `hue_factor` is the amount of shift in H channel and must be in the
    interval `[-0.5, 0.5]`.

    See `Hue`_ for more details.

    .. _Hue: https://en.wikipedia.org/wiki/Hue

    Args::

        [in] img (PIL Image.Image): Image to be adjusted.
        [in] hue_factor (float):  How much to shift the hue channel.
             Should be in [-0.5, 0.5]. 0.5 and -0.5 give complete reversal of
             hue channel in HSV space in positive and negative direction respectively.
             0 means no shift. Therefore, both -0.5 and 0.5 will give an image
             with complementary colors while 0 gives the original image.

    Returns::

        [out] PIL Image.Image: Saturation adjusted image.

    Example::
        
        img = Image.open(...)
        img_ = transform.adjust_hue(img, 0.1)
    """
    if not(-0.5 <= hue_factor <= 0.5):
        raise ValueError(f'hue_factor ({hue_factor}) is not in [-0.5, 0.5].')

    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    input_mode = img.mode
    if input_mode in {'L', '1', 'I', 'F'}:
        return img

    h, s, v = img.convert('HSV').split()

    np_h = np.array(h, dtype=np.uint8)
    # uint8 addition take cares of rotation across boundaries
    with np.errstate(over='ignore'):
        np_h += np.uint8(hue_factor * 255)
    h = Image.fromarray(np_h, 'L')

    img = Image.merge('HSV', (h, s, v)).convert(input_mode)
    return img


def adjust_gamma(img, gamma, gain=1):
    """
    Function for performing gamma correction on an image.

    Also known as Power Law Transform. Intensities in RGB mode are adjusted
    based on the following equation:

    .. math::
        I_{\text{out}} = 255 \times \text{gain} \times \left(\frac{I_{\text{in}}}{255}\right)^{\gamma}

    See `Gamma Correction`_ for more details.

    .. _Gamma Correction: https://en.wikipedia.org/wiki/Gamma_correction

    Args::

        [in] img (PIL Image.Image): Image to be adjusted.
        [in] gamma (float): Non negative real number, same as :math:`\gamma` in the equation.
             gamma larger than 1 make the shadows darker,
             while gamma smaller than 1 make dark regions lighter.
        [in] gain (float): The constant multiplier.

    Returns::

        [out] PIL Image.Image: Gamma adjusted image.
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    if gamma < 0:
        raise ValueError('Gamma should be a non-negative real number')

    input_mode = img.mode
    img = img.convert('RGB')
    gamma_map = [(255 + 1 - 1e-3) * gain * pow(ele / 255., gamma) for ele in range(256)] * 3
    img = img.point(gamma_map)  # use PIL's point-function to accelerate this part

    img = img.convert(input_mode)
    return img


def crop(img, top, left, height, width):
    """
    Function for cropping image.

    Args::

        [in] img(PIL Image.Image): Input image.
        [in] top(int): the top boundary of the cropping box.
        [in] left(int): the left boundary of the cropping box.
        [in] height(int): height of the cropping box.
        [in] width(int): width of the cropping box.

    Returns::

        [out] PIL Image.Image: Cropped image.

    Example::
        
        img = Image.open(...)
        img_ = transform.crop(img, 10, 10, 100, 100)
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    return img.crop((left, top, left + width, top + height))


def resize(img, size, interpolation=Image.BILINEAR):
    """
    Function for resizing the input image to the given size.

    Args::

        [in] img(PIL Image.Image): Input image.
        [in] size(sequence or int): Desired output size. If size is a sequence like
             (h, w), the output size will be matched to this. If size is an int,
             the smaller edge of the image will be matched to this number maintaining
             the aspect ratio. If a tuple or list of length 1 is provided, it is
             interpreted as a single int.
        [in] interpolation(int, optional): type of interpolation. default: PIL.Image.BILINEAR

    Returns::

        [out] PIL Image.Image: Resized image.

    Example::
        
        img = Image.open(...)
        img_ = transform.resize(img, (100, 100))
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')
    if not (isinstance(size, int) or (isinstance(size, Sequence) and len(size) in (1, 2))):
        raise TypeError(f'Got inappropriate size arg: {size}')

    if isinstance(size, int) or len(size) == 1:
        if isinstance(size, Sequence):
            size = size[0]
        w, h = img.size
        if (w <= h and w == size) or (h <= w and h == size):
            return img
        if w < h:
            ow = size
            oh = int(size * h / w)
            return img.resize((ow, oh), interpolation)
        else:
            oh = size
            ow = int(size * w / h)
            return img.resize((ow, oh), interpolation)
    else:
        return img.resize(size[::-1], interpolation)


def gray(img, num_output_channels):
    """
    Function for converting PIL image of any mode (RGB, HSV, LAB, etc) to grayscale version of image.

    Args::

        [in] img(PIL Image.Image): Input image.
        [in] num_output_channels (int): number of channels of the output image. Value can be 1 or 3. Default, 1.

    Returns::

        [out] PIL Image: Grayscale version of the image.
              if num_output_channels = 1 : returned image is single channel
              if num_output_channels = 3 : returned image is 3 channel with r = g = b
    """
    if not _is_pil_image(img):
        raise TypeError(f'img should be PIL Image. Got {type(img)}')

    if num_output_channels == 1:
        img = img.convert('L')
    elif num_output_channels == 3:
        img = img.convert('L')
        np_img = np.array(img, dtype=np.uint8)
        np_img = np.dstack([np_img, np_img, np_img])
        img = Image.fromarray(np_img, 'RGB')
    else:
        raise ValueError('num_output_channels should be either 1 or 3')

    return img

def _get_perspective_coeffs(startpoints, endpoints):
    """Helper function to get the coefficients (a, b, c, d, e, f, g, h) for the perspective transforms.

    In Perspective Transform each pixel (x, y) in the orignal image gets transformed as,
     (x, y) -> ( (ax + by + c) / (gx + hy + 1), (dx + ey + f) / (gx + hy + 1) )

    Args:
        List containing [top-left, top-right, bottom-right, bottom-left] of the orignal image,
        List containing [top-left, top-right, bottom-right, bottom-left] of the transformed
                   image
    Returns:
        octuple (a, b, c, d, e, f, g, h) for transforming each pixel.
    """
    matrix = []

    for p1, p2 in zip(endpoints, startpoints):
        matrix.append([p1[0], p1[1], 1, 0, 0, 0, -p2[0] * p1[0], -p2[0] * p1[1]])
        matrix.append([0, 0, 0, p1[0], p1[1], 1, -p2[1] * p1[0], -p2[1] * p1[1]])

    A = np.array(matrix, dtype="float")
    B = np.array(startpoints, dtype="float").reshape(8)
    res = np.linalg.lstsq(A, B, rcond=-1)[0]
    return res.tolist()


def perspective(img, startpoints, endpoints, interpolation=Image.BICUBIC):
    """Perform perspective transform of the given PIL Image.

    Args:
        img (PIL Image): Image to be transformed.
        startpoints: List containing [top-left, top-right, bottom-right, bottom-left] of the orignal image
        endpoints: List containing [top-left, top-right, bottom-right, bottom-left] of the transformed image
        interpolation: Default- Image.BICUBIC
    Returns:
        PIL Image:  Perspectively transformed Image.
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    coeffs = _get_perspective_coeffs(startpoints, endpoints)
    return img.transform(img.size, Image.PERSPECTIVE, coeffs, interpolation)


def resized_crop(img, top, left, height, width, size, interpolation=Image.BILINEAR):
    """Crop the given PIL Image and resize it to desired size.

    Notably used in :class:`~torchvision.transforms.RandomResizedCrop`.

    Args:
        img (PIL Image): Image to be cropped. (0,0) denotes the top left corner of the image.
        top (int): Vertical component of the top left corner of the crop box.
        left (int): Horizontal component of the top left corner of the crop box.
        height (int): Height of the crop box.
        width (int): Width of the crop box.
        size (sequence or int): Desired output size. Same semantics as ``resize``.
        interpolation (int, optional): Desired interpolation. Default is
            ``PIL.Image.BILINEAR``.
    Returns:
        PIL Image: Cropped image.
    """
    assert _is_pil_image(img), 'img should be PIL Image'
    img = crop(img, top, left, height, width)
    img = resize(img, size, interpolation)
    return img

def center_crop(img, output_size):
    """Crop the given PIL Image and resize it to desired size.

        Args:
            img (PIL Image): Image to be cropped. (0,0) denotes the top left corner of the image.
            output_size (sequence or int): (height, width) of the crop box. If int,
                it is used for both directions
        Returns:
            PIL Image: Cropped image.
        """
    if isinstance(output_size, numbers.Number):
        output_size = (int(output_size), int(output_size))
    image_width, image_height = img.size
    crop_height, crop_width = output_size
    crop_top = int(round((image_height - crop_height) / 2.))
    crop_left = int(round((image_width - crop_width) / 2.))
    return crop(img, crop_top, crop_left, crop_height, crop_width)

def five_crop(img, size):
    """Crop the given PIL Image into four corners and the central crop.

    .. Note::
        This transform returns a tuple of images and there may be a
        mismatch in the number of inputs and targets your ``Dataset`` returns.

    Args:
       size (sequence or int): Desired output size of the crop. If size is an
           int instead of sequence like (h, w), a square crop (size, size) is
           made.

    Returns:
       tuple: tuple (tl, tr, bl, br, center)
                Corresponding top left, top right, bottom left, bottom right and center crop.
    """
    if isinstance(size, numbers.Number):
        size = (int(size), int(size))
    else:
        assert len(size) == 2, "Please provide only two dimensions (h, w) for size."

    image_width, image_height = img.size
    crop_height, crop_width = size
    if crop_width > image_width or crop_height > image_height:
        msg = "Requested crop size {} is bigger than input size {}"
        raise ValueError(msg.format(size, (image_height, image_width)))

    tl = img.crop((0, 0, crop_width, crop_height))
    tr = img.crop((image_width - crop_width, 0, image_width, crop_height))
    bl = img.crop((0, image_height - crop_height, crop_width, image_height))
    br = img.crop((image_width - crop_width, image_height - crop_height,
                   image_width, image_height))
    center = center_crop(img, (crop_height, crop_width))
    return (tl, tr, bl, br, center)

def ten_crop(img, size, vertical_flip=False):
    r"""Crop the given PIL Image into four corners and the central crop plus the
        flipped version of these (horizontal flipping is used by default).

    .. Note::
        This transform returns a tuple of images and there may be a
        mismatch in the number of inputs and targets your ``Dataset`` returns.

    Args:
       size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
       vertical_flip (bool): Use vertical flipping instead of horizontal

    Returns:
       tuple: tuple (tl, tr, bl, br, center, tl_flip, tr_flip, bl_flip, br_flip, center_flip)
                Corresponding top left, top right, bottom left, bottom right and center crop
                and same for the flipped image.
    """
    if isinstance(size, numbers.Number):
        size = (int(size), int(size))
    else:
        assert len(size) == 2, "Please provide only two dimensions (h, w) for size."

    first_five = five_crop(img, size)

    if vertical_flip:
        img = vflip(img)
    else:
        img = hflip(img)

    second_five = five_crop(img, size)
    return first_five + second_five


def rotate(img, angle, resample=False, expand=False, center=None, fill=None):
    """Rotate the image by angle.


    Args:
        img (PIL Image): PIL Image to be rotated.
        angle (float or int): In degrees degrees counter clockwise order.
        resample (``PIL.Image.NEAREST`` or ``PIL.Image.BILINEAR`` or ``PIL.Image.BICUBIC``, optional):
            An optional resampling filter. See `filters`_ for more information.
            If omitted, or if the image has mode "1" or "P", it is set to ``PIL.Image.NEAREST``.
        expand (bool, optional): Optional expansion flag.
            If true, expands the output image to make it large enough to hold the entire rotated image.
            If false or omitted, make the output image the same size as the input image.
            Note that the expand flag assumes rotation around the center and no translation.
        center (2-tuple, optional): Optional center of rotation.
            Origin is the upper left corner.
            Default is the center of the image.
        fill (n-tuple or int or float): Pixel fill value for area outside the rotated
            image. If int or float, the value is used for all bands respectively.
            Defaults to 0 for all bands. This option is only available for ``pillow>=5.2.0``.

    .. _filters: https://pillow.readthedocs.io/en/latest/handbook/concepts.html#filters

    """
    def parse_fill(fill, num_bands):
        if PILLOW_VERSION < "5.2.0":
            if fill is None:
                return {}
            else:
                msg = ("The option to fill background area of the rotated image, "
                       "requires pillow>=5.2.0")
                raise RuntimeError(msg)

        if fill is None:
            fill = 0
        if isinstance(fill, (int, float)) and num_bands > 1:
            fill = tuple([fill] * num_bands)
        if not isinstance(fill, (int, float)) and len(fill) != num_bands:
            msg = ("The number of elements in 'fill' does not match the number of "
                   "bands of the image ({} != {})")
            raise ValueError(msg.format(len(fill), num_bands))

        return {"fillcolor": fill}

    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    opts = parse_fill(fill, len(img.getbands()))

    return img.rotate(angle, resample, expand, center, **opts)


def _get_inverse_affine_matrix(center, angle, translate, scale, shear):
    # Helper method to compute inverse matrix for affine transformation

    # As it is explained in PIL.Image.rotate
    # We need compute INVERSE of affine transformation matrix: M = T * C * RSS * C^-1
    # where T is translation matrix: [1, 0, tx | 0, 1, ty | 0, 0, 1]
    #       C is translation matrix to keep center: [1, 0, cx | 0, 1, cy | 0, 0, 1]
    #       RSS is rotation with scale and shear matrix
    #       RSS(a, s, (sx, sy)) =
    #       = R(a) * S(s) * SHy(sy) * SHx(sx)
    #       = [ s*cos(a - sy)/cos(sy), s*(-cos(a - sy)*tan(x)/cos(y) - sin(a)), 0 ]
    #         [ s*sin(a + sy)/cos(sy), s*(-sin(a - sy)*tan(x)/cos(y) + cos(a)), 0 ]
    #         [ 0                    , 0                                      , 1 ]
    #
    # where R is a rotation matrix, S is a scaling matrix, and SHx and SHy are the shears:
    # SHx(s) = [1, -tan(s)] and SHy(s) = [1      , 0]
    #          [0, 1      ]              [-tan(s), 1]
    #
    # Thus, the inverse is M^-1 = C * RSS^-1 * C^-1 * T^-1

    if isinstance(shear, numbers.Number):
        shear = [shear, 0]

    if not isinstance(shear, (tuple, list)) and len(shear) == 2:
        raise ValueError(
            "Shear should be a single value or a tuple/list containing " +
            "two values. Got {}".format(shear))

    rot = math.radians(angle)
    sx, sy = [math.radians(s) for s in shear]

    cx, cy = center
    tx, ty = translate

    # RSS without scaling
    a = cos(rot - sy) / cos(sy)
    b = -cos(rot - sy) * tan(sx) / cos(sy) - sin(rot)
    c = sin(rot - sy) / cos(sy)
    d = -sin(rot - sy) * tan(sx) / cos(sy) + cos(rot)

    # Inverted rotation matrix with scale and shear
    # det([[a, b], [c, d]]) == 1, since det(rotation) = 1 and det(shear) = 1
    M = [d, -b, 0,
         -c, a, 0]
    M = [x / scale for x in M]

    # Apply inverse of translation and of center translation: RSS^-1 * C^-1 * T^-1
    M[2] += M[0] * (-cx - tx) + M[1] * (-cy - ty)
    M[5] += M[3] * (-cx - tx) + M[4] * (-cy - ty)

    # Apply center translation: C * RSS^-1 * C^-1 * T^-1
    M[2] += cx
    M[5] += cy
    return M


def affine(img, angle, translate, scale, shear, resample=0, fillcolor=None):
    """Apply affine transformation on the image keeping image center invariant

    Args:
        img (PIL Image): PIL Image to be rotated.
        angle (float or int): rotation angle in degrees between -180 and 180, clockwise direction.
        translate (list or tuple of integers): horizontal and vertical translations (post-rotation translation)
        scale (float): overall scale
        shear (float or tuple or list): shear angle value in degrees between -180 to 180, clockwise direction.
        If a tuple of list is specified, the first value corresponds to a shear parallel to the x axis, while
        the second value corresponds to a shear parallel to the y axis.
        resample (``PIL.Image.NEAREST`` or ``PIL.Image.BILINEAR`` or ``PIL.Image.BICUBIC``, optional):
            An optional resampling filter.
            See `filters`_ for more information.
            If omitted, or if the image has mode "1" or "P", it is set to ``PIL.Image.NEAREST``.
        fillcolor (int): Optional fill color for the area outside the transform in the output image. (Pillow>=5.0.0)
    """
    if not _is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    assert isinstance(translate, (tuple, list)) and len(translate) == 2, \
        "Argument translate should be a list or tuple of length 2"

    assert scale > 0.0, "Argument scale should be positive"

    output_size = img.size
    center = (img.size[0] * 0.5 + 0.5, img.size[1] * 0.5 + 0.5)
    matrix = _get_inverse_affine_matrix(center, angle, translate, scale, shear)
    kwargs = {"fillcolor": fillcolor} if PILLOW_VERSION[0] >= '5' else {}
    return img.transform(output_size, Image.AFFINE, matrix, resample, **kwargs)




# Author: Zheng-Ning Liu 
# 
# The gpu implementation is original provided by Haoqiang Fan and Kaichun Mo,
# <https://github.com/daerduoCarey/PyTorchEMD>.

import jittor as jt
from jittor import Function

EMD_gpu_header = '''
namespace jittor {
__device__ inline out_type dist2(out_type x1, out_type y1, out_type z1,
        out_type x2, out_type y2, out_type z2) {
    return (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1) + (z2 - z1) * (z2 - z1);
}
}
'''

approxmatch_gpu_src = '''
    __global__ void approxmatch_gpu_kernel(@ARGS_DEF) {
        @PRECALC
        @alias(xyz1, in0)
        @alias(xyz2, in1)
        @alias(match, out)

        int b = in0_shape0;
        int n = in0_shape1;
        int m = in1_shape1;

        out_type *remainL = in2_p + blockIdx.x * (n + m) * 2;
        out_type *remainR = remainL + n;
        out_type *ratioL = remainR + m;
        out_type *ratioR = ratioL + n;

        const int Block = 1024;
        __shared__ out_type buf[Block * 4];

        for (int i = blockIdx.x; i < b; i += gridDim.x) {
            for (int j = threadIdx.x; j < n * m; j += blockDim.x)
                match_p[i * n * m + j] = 0;
            for (int j = threadIdx.x; j < n; j += blockDim.x)
                remainL[j] = n >= m ? 1 : m / n;
            for (int j = threadIdx.x; j < m; j += blockDim.x)
                remainR[j] = n >= m ? n / m : 1;
            __syncthreads();

            for (int j = 7; j >= -2; j--) {
                out_type level = j > -2 ? -powf(4.0f, j) : 0;

                for (int k0 = 0; k0 < n; k0 += blockDim.x) {
                    int k = k0 + threadIdx.x;
                    out_type x1 = 0, y1 = 0, z1 = 0;
                    if (k < n) {
                        x1 = @xyz1(i, k, 0);
                        y1 = @xyz1(i, k, 1);
                        z1 = @xyz1(i, k, 2);
                    }

                    out_type suml = 1e-9f;
                    for (int l0 = 0; l0 < m; l0 += Block){
                        int lend = min(m, l0 + Block) - l0;
                        for (int l = threadIdx.x; l < lend; l += blockDim.x) {
                            buf[l * 4 + 0] = @xyz2(i, l0 + l, 0);
                            buf[l * 4 + 1] = @xyz2(i, l0 + l, 1);
                            buf[l * 4 + 2] = @xyz2(i, l0 + l, 2);
                            buf[l * 4 + 3] = remainR[l0 + l];
                        }
                        __syncthreads();

                        for (int l = 0; l < lend; l++){
                            out_type x2 = buf[l * 4 + 0];
                            out_type y2 = buf[l * 4 + 1];
                            out_type z2 = buf[l * 4 + 2];
                            out_type d = level * dist2(x1, y1, z1, x2, y2, z2);
                            out_type w = __expf(d) * buf[l * 4 + 3];
                            suml += w;
                        }
                        __syncthreads();
                    }
                    if (k < n)
                        ratioL[k] = remainL[k] / suml;
                }
                __syncthreads();

                for (int l0 = 0; l0 < m; l0 += blockDim.x){
                    int l = l0 + threadIdx.x;
                    out_type x2 = 0, y2 = 0, z2 = 0;
                    if (l < m){
                        x2 = @xyz2(i, l, 0);
                        y2 = @xyz2(i, l, 1);
                        z2 = @xyz2(i, l, 2);
                    }
                    out_type sumr = 0;
                    for (int k0 = 0; k0 < n; k0 += Block){
                        int kend = min(n, k0 + Block) - k0;
                        for (int k = threadIdx.x; k < kend; k += blockDim.x){
                            buf[k * 4 + 0] = @xyz1(i, k0 + k, 0);
                            buf[k * 4 + 1] = @xyz1(i, k0 + k, 1);
                            buf[k * 4 + 2] = @xyz1(i, k0 + k, 2);
                            buf[k * 4 + 3] = ratioL[k0 + k];
                        }
                        __syncthreads();

                        for (int k = 0; k < kend; k++){
                            out_type x1 = buf[k * 4 + 0];
                            out_type y1 = buf[k * 4 + 1];
                            out_type z1 = buf[k * 4 + 2];
                            out_type d = level * dist2(x1, y1, z1, x2, y2, z2);
                            out_type w = __expf(d) * buf[k * 4 + 3];
                            sumr += w;
                        }
                        __syncthreads();
                    }

                    if (l < m){
                        sumr *= remainR[l];
                        out_type consumption = fminf(remainR[l] / (sumr + 1e-9f), 1.0f);
                        ratioR[l] = consumption * remainR[l];
                        remainR[l] = fmaxf(0.0f, remainR[l] - sumr);
                    }
                }
                __syncthreads();

                for (int k0 = 0; k0 < n; k0 += blockDim.x){
                    int k = k0 + threadIdx.x;
                    out_type x1 = 0, y1 = 0, z1 = 0;
                    if (k < n){
                        x1 = @xyz1(i, k, 0);
                        y1 = @xyz1(i, k, 1);
                        z1 = @xyz1(i, k, 2);
                    }
                    out_type suml = 0;
                    for (int l0 = 0; l0 < m; l0 += Block){
                        int lend = min(m, l0 + Block)-l0;
                        for (int l = threadIdx.x; l < lend; l += blockDim.x){
                            buf[l * 4 + 0] = @xyz2(i, l0 + l, 0);
                            buf[l * 4 + 1] = @xyz2(i, l0 + l, 1);
                            buf[l * 4 + 2] = @xyz2(i, l0 + l, 2);
                            buf[l * 4 + 3] = ratioR[l0 + l];
                        }
                        __syncthreads();

                        out_type rl = ratioL[k];
                        if (k < n){
                            for (int l = 0; l < lend; l++){
                                out_type x2 = buf[l * 4 + 0];
                                out_type y2 = buf[l * 4 + 1];
                                out_type z2 = buf[l * 4 + 2];
                                out_type d = level * dist2(x1, y1, z1, x2, y2, z2);
                                out_type w = __expf(d) * rl * buf[l*4+3];
                                @match(i, l0 + l, k) += w;
                                suml += w;
                            }
                        }
                        __syncthreads();
                    }
                    if (k < n)
                        remainL[k] = fmaxf(0.0f, remainL[k] - suml);
                }
                __syncthreads();
            }
        }
    }

    approxmatch_gpu_kernel<<<32, 512>>>(@ARGS);
'''

matchcost_gpu_src = '''
    __global__ void matchcost_gpu_kernel(@ARGS_DEF) {
        @PRECALC
        @alias(xyz1, in0)
        @alias(xyz2, in1)
        @alias(match, in2)

        int b = in0_shape0;
        int n = in0_shape1;
        int m = in1_shape1;

        const int Block = 1024;
        __shared__ out_type allsum[512];
        __shared__ out_type buf[Block * 3];

        for (int i = blockIdx.x; i < b; i += gridDim.x) {
            out_type subsum = 0;
            for (int k0 = 0; k0 < n; k0 += blockDim.x) {
                int k = k0 + threadIdx.x;
                out_type x1 = 0, y1 = 0, z1 = 0;
                if (k < n) {
                    x1 = @xyz1(i, k, 0);
                    y1 = @xyz1(i, k, 1);
                    z1 = @xyz1(i, k, 2);
                }

                for (int l0 = 0; l0 < m; l0 += Block) {
                    int lend = min(m, l0 + Block) - l0;
                    for (int l = threadIdx.x; l < lend * 3; l += blockDim.x)
                        buf[l] = xyz2_p[i * m * 3 + l0 * 3 + l];
                    __syncthreads();

                    if (k < n) {
                        for (int l = 0; l < lend; l++) {
                            out_type x2 = buf[l * 3 + 0];
                            out_type y2 = buf[l * 3 + 1];
                            out_type z2 = buf[l * 3 + 2];
                            out_type d = dist2(x1, y1, z1, x2, y2, z2);
                            subsum += d * @match(i, l0 + l, k);
                        }
                    }
                    __syncthreads();
                }
            }

            allsum[threadIdx.x] = subsum;
            for (int j = 1; j < blockDim.x; j <<= 1) {
                __syncthreads();
                if ((threadIdx.x & j) == 0 && threadIdx.x + j < blockDim.x) {
                    allsum[threadIdx.x] += allsum[threadIdx.x + j];
                }
            }

            if (threadIdx.x == 0)
                @out(i) = allsum[0];
            __syncthreads();
        }
    }

    matchcost_gpu_kernel<<<32, 512>>>(@ARGS);
'''

matchcost_grad1_gpu_src = '''
    __global__ void matchcost_grad1_gpu_kernel(@ARGS_DEF) {
        @PRECALC
        @alias(grad, in0)
        @alias(xyz1, in1)
        @alias(xyz2, in2)
        @alias(match, in3)

        int b = grad_shape0;
        int n = xyz1_shape1;
        int m = xyz2_shape1;

        for (int i = blockIdx.x; i < b ; i += gridDim.x){
            for (int l = threadIdx.x; l < n; l += blockDim.x){
                out_type x1 = @xyz1(i, l, 0);
                out_type y1 = @xyz1(i, l, 1);
                out_type z1 = @xyz1(i, l, 2);
                out_type dx = 0, dy = 0, dz = 0;
                for (int k = 0; k < m; k++){
                    out_type x2 = @xyz2(i, k, 0);
                    out_type y2 = @xyz2(i, k, 1);
                    out_type z2 = @xyz2(i, k, 2);
                    out_type d = @match(i, k, l) * 2;
                    dx += (x1 - x2) * d;
                    dy += (y1 - y2) * d;
                    dz += (z1 - z2) * d;
                }
                @out(i, l, 0) = dx * @grad(i);
                @out(i, l, 1) = dy * @grad(i);
                @out(i, l, 2) = dz * @grad(i);
            }
        }
    }

    matchcost_grad1_gpu_kernel<<<32, 512>>>(@ARGS);
'''

matchcost_grad2_gpu_src = '''
    __global__ void matchcost_grad2_gpu_kernel(@ARGS_DEF) {
        @PRECALC
        @alias(grad, in0)
        @alias(xyz1, in1)
        @alias(xyz2, in2)
        @alias(match, in3)

        int b = grad_shape0;
        int n = xyz1_shape1;
        int m = xyz2_shape1;

        __shared__ out_type sum_grad[256 * 3];
        for (int i = blockIdx.x; i < b; i += gridDim.x) {
            int kbeg = m * blockIdx.y / gridDim.y;
            int kend = m * (blockIdx.y + 1) / gridDim.y;
            for (int k = kbeg; k < kend; k++) {
                out_type x2 = @xyz2(i, k, 0);
                out_type y2 = @xyz2(i, k, 1);
                out_type z2 = @xyz2(i, k, 2);
                out_type subsumx = 0, subsumy = 0, subsumz = 0;
                for (int j = threadIdx.x; j < n; j += blockDim.x) {
                    out_type x1 = x2 - @xyz1(i, j, 0);
                    out_type y1 = y2 - @xyz1(i, j, 1);
                    out_type z1 = z2 - @xyz1(i, j, 2);
                    out_type d = @match(i, k, j) * 2;
                    subsumx += x1 * d;
                    subsumy += y1 * d;
                    subsumz += z1 * d;
                }
                sum_grad[threadIdx.x * 3 + 0] = subsumx;
                sum_grad[threadIdx.x * 3 + 1] = subsumy;
                sum_grad[threadIdx.x * 3 + 2] = subsumz;

                for (int j = 1; j < blockDim.x; j <<= 1) {
                    __syncthreads();
                    int j1 = threadIdx.x;
                    int j2 = threadIdx.x + j;
                    if ((j1 & j) == 0 && j2 < blockDim.x){
                        sum_grad[j1 * 3 + 0] += sum_grad[j2 * 3 + 0];
                        sum_grad[j1 * 3 + 1] += sum_grad[j2 * 3 + 1];
                        sum_grad[j1 * 3 + 2] += sum_grad[j2 * 3 + 2];
                    }
                }
                if (threadIdx.x == 0){
                    @out(i, k, 0) = sum_grad[0] * @grad(i);
                    @out(i, k, 1) = sum_grad[1] * @grad(i);
                    @out(i, k, 2) = sum_grad[2] * @grad(i);
                }
                __syncthreads();
            }
        }
    }

    matchcost_grad2_gpu_kernel<<<dim3(32, 32), 256>>>(@ARGS);
'''

class EarthMoverDistance(Function):
    ''' A loss layer that computes Earth Mover's distance from pc1 to pc2. Only supports GPU.

    :param pc1:  input point cloud
    :type pc1: jittor array

    :param pc2:  input point cloud
    :type pc2: jittor array

    :param reduction: reduction method in batches, can be 'mean', 'sum', or None. Default: 'mean'.
    :type reduction: str, optional
            
    :param dims: a string that represents each dimension, can be
            '[BNC]' ([batch, number of points, xyz]), or
            '[BCN]' ([batch, xyz, number of points]). Default: 'BNC'.
    :type dims: str, optional

    Example:

    >>> import jittor as jt
    >>> from jittor.loss3d import EarthMoverDistance
    >>> jt.flags.use_cuda = True
    >>> pc1 = jt.rand([10, 100, 3], dtype=jt.float32)
    >>> pc2 = jt.rand([10, 100, 3], dtype=jt.float32)
    >>> EMD = EarthMoverDistance(dims='BNC')
    >>> emd = EMD(pc1, pc2)
    >>> print('EMD =', emd.item())
    '''
    def execute(self, pc1, pc2, reduction='mean', dims='BNC'):
        assert dims in ['BNC', 'BCN']
        if dims == 'BCN':
            pc1, pc2 = pc1.permute(0, 2, 1), pc2.permute(0, 2, 1)

        batch_size_1, N, _ = pc1.shape
        batch_size_2, M, _ = pc2.shape
        assert batch_size_1 == batch_size_2
        batch_size = batch_size_1

        temp = jt.zeros([batch_size, (N + M) * 2], pc1.dtype)
        match = jt.code(
            shape=[batch_size, M, N],
            dtype=pc1.dtype,
            inputs=[pc1, pc2, temp],
            cuda_header=EMD_gpu_header,
            cuda_src=approxmatch_gpu_src,
        )

        emd = jt.code(
            shape=[batch_size],
            dtype=pc1.dtype,
            inputs=[pc1, pc2, match],
            cuda_header=EMD_gpu_header,
            cuda_src=matchcost_gpu_src,
        )

        self.saved_vars = (pc1, pc2, match, reduction)

        if reduction is None:
            return emd
        elif reduction == 'sum':
            return emd.sum()
        elif reduction == 'mean':
            return emd.mean()

    def grad(self, grad):
        pc1, pc2, match, reduction = self.saved_vars

        if reduction == 'sum':
            grad = jt.ones([pc1.shape[0]]) * grad
        elif reduction == 'mean':
            grad = jt.ones([pc1.shape[0]]) * grad / pc1.shape[0]

        grad_pc1 = jt.code(
            shape=pc1.shape,
            dtype=pc1.dtype,
            inputs=[grad, pc1, pc2, match],
            cuda_src=matchcost_grad1_gpu_src,
        )

        grad_pc2 = jt.code(
            shape=pc2.shape,
            dtype=pc2.dtype,
            inputs=[grad, pc1, pc2, match],
            cuda_src=matchcost_grad2_gpu_src,
        )

        return grad_pc1, grad_pc2


def earth_mover_distance(pc1, pc2, reduction='mean', dims='BNC'):
    ''' Earth Mover's distance from pc1 to pc2. Only supports GPU.

    :param pc1:  input point cloud
    :type pc1: jittor array

    :param pc2:  input point cloud
    :type pc2: jittor array

    :param reduction: reduction method in batches, can be 'mean', 'sum', or None. Default: 'mean'.
    :type reduction: str, optional
            
    :param dims: a string that represents each dimension, can be
            '[BNC]' ([batch, number of points, xyz]), or
            '[BCN]' ([batch, xyz, number of points]). Default: 'BNC'.
    :type dims: str, optional


    Example:

    >>> import jittor as jt
    >>> from jittor.loss3d import earth_mover_distance
    >>> jt.flags.use_cuda = True
    >>> pc1 = jt.rand([10, 100, 3], dtype=jt.float32)
    >>> pc2 = jt.rand([10, 100, 3], dtype=jt.float32)
    >>> emd = earth_mover_distance(pc1, pc2, dims='BNC')
    >>> print('EMD =', emd.item())
    '''
    return EarthMoverDistance.apply(pc1, pc2, reduction, dims)



# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers:
#   Meng-Hao Guo <guomenghao1997@gmail.com>
#   Dun Liang <randonlang@gmail.com>.
#
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import jittor as jt
import unittest
import sys, os
from subprocess import getoutput

class TestLazyExecution(unittest.TestCase):
    @unittest.skipIf(not jt.has_cuda, "No cuda found")
    def test_lazy_execution(self):
        code = """
import jittor as jt
jt.flags.use_cuda = 1

a = jt.zeros(1)
b = jt.code([1], a.dtype, [a],
cuda_header='''
#include <assert.h>
''',
cuda_src='''
__global__ void kernel(float32* a, float32* b) {
    b[0] = a[0];
    assert(a[0] == 1);
}
kernel<<<1,1>>>(in0_p, out0_p);
''')
c = a+b
print(c)
"""
        fpath = os.path.join(jt.flags.cache_path, "lazy_error.py")
        with open(fpath, 'w') as f:
            f.write(code)
        res = getoutput(f"{sys.executable} {fpath}")
        assert 'print(c)' in res
        res = getoutput(f"lazy_execution=0 {sys.executable} {fpath}")
        assert "''')" in res
        


if __name__ == "__main__":
    unittest.main()




# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: 
#     Wenyang Zhou <576825820@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
import jittor.nn as jnn

skip_this_test = False

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    import torch.nn as tnn
    import torchvision
except:
    torch = None
    tnn = None
    torchvision = None
    skip_this_test = True

@unittest.skipIf(skip_this_test, "No Torch found")
class TestHook(unittest.TestCase):
    def test_bhook(self):
        a = jnn.ReLU()
        hooked = False
        def hook(mod, grad_input, grad_output):
            nonlocal hooked
            hooked = True
            assert len(grad_input) == 1
            assert len(grad_output) == 1
            np.testing.assert_allclose(grad_input[0].numpy(), [0, 1])
            np.testing.assert_allclose(grad_output[0].numpy(), [1, 1])
            return (jt.array([-1.0, -2.0]), )
        a.register_backward_hook(hook)
        x = jt.array([-1.0,2])
        y = a(x)
        dx = jt.grad(y, x)
        assert hooked
        np.testing.assert_allclose(dx.numpy(), [-1.0, -2.0])

    def test_register_hook(self):
        x = jt.array([0.0, 0.0])
        y = x * [1,2]
        y.register_hook(lambda g: g*2)
        dx = jt.grad(y, x)
        np.testing.assert_allclose(dx.data, [2,4])

    def test_requires_grads_(self):
        class Mod(jt.nn.Module):
            def execute(self, x):
                return x*2
        x = jt.random((100,))
        mod = Mod()
        mod.requires_grad_(True)
        y = mod(x)
        y = y*10
        dx = jt.grad(y, x)
        np.testing.assert_allclose(dx.data, 20)

        mod.requires_grad_(False)
        y = mod(x)
        y = y*10
        dx = jt.grad(y, x)
        np.testing.assert_allclose(dx.data, 0)

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
from jittor import LOG
import numpy as np
from .test_core import expect_error

jit_eval = jt.core.op_compiler.eval
jit_precompile = jt.core.op_compiler.precompile

class TestOpCompiler(unittest.TestCase):
    def test_eval(self):
        def check(expr, vars={}):
            for k,v in vars.items():
                locals()[k] = int(v)
            _v1 = None
            _v2 = None
            try:
                _v1 = jit_eval(expr, vars)
            except:
                pass
            try:
                _v2 = eval(expr)
            except:
                pass
            LOG.vv(f"check {expr} = {_v1}, {_v2}, {_v1 == _v2}")
            assert _v1 == _v2
        check("10+2*6")
        check("100 * 2 + 12")
        check("100*2+12")
        check("100 * ( 2 + 12 )")
        check("100*(2+12)")
        check("100 * ( 2 + 12 ) / 14")
        check("100*(2+12)/14")
        check("-1")
        check("- 1")
        vars = {"a":"123", "b":"2"}
        check("a", vars)
        check("a+b", vars)
        # python divide is different with c++
        # check("a/b", vars)
        check("-1 +a *b", vars)
        check("*****", vars)
        
    def test_precompile_ifdef(self):
        vars = {"JIT_a":"1"}
        check = lambda expr, result: \
            self.assertEqual(jit_precompile(vars, expr), result)
        check("#ifdef JIT_a\nxxx\n#endif", "xxx\n")
        check("#ifdef JIT_a\nxxx\n#else\nyyy\n #endif", "xxx\n")
        check("#ifndef JIT_a\nxxx\n#else\nyyy\n #endif", "yyy\n ")
        check("#ifdef JIT_b\nxxx\n#else\nyyy\n #endif", "yyy\n ")
        check("#ifdef b\nxxx\n#else\nyyy\n #endif",
              "#ifdef b\nxxx\n#else\nyyy\n #endif")
        for va in [0,1]:
            for vb in [0,1]:
                vars["JIT_a"] = "1"
                vars["JIT_b"] = "1"
                if not va:  del vars["JIT_a"]
                if not vb:  del vars["JIT_b"]
                check((
                    "#ifdef JIT_a\n"
                    "#ifdef JIT_b\n"
                    "0\n"
                    "#else\n"
                    "1\n"
                    "#endif\n"
                    "#else\n"
                    "#ifdef JIT_b\n"
                    "2\n"
                    "#else\n"
                    "3\n"
                    "#endif\n"
                    "#endif\n"
                ), f"{3 - (va*2+vb)}\n")
        
    def test_precompile(self):
        vars = {"a":"2", "b":"5", "a1":"1", "a2":"2", "OP":"mean"}
        check = lambda expr, result: \
            self.assertEqual(jit_precompile(vars, expr), result) 
        check("@", "@")
        check("@a", "2")
        # check("//@a\n@a", "//@a\n2")
        check("//@a\n@a", "\n2")
        # check("@a//@a", "2//@a")
        check("@a//@a", "2")
        check("@{-a +b* 2}", "8")
        # check("@{-a +b* 2}/*@{-a +b* 2}*/", "8/*@{-a +b* 2}*/")
        check("@{-a +b* 2}/*@{-a +b* 2}*/", "8")
        check("@for(i,a,b,+@i)", "+2+3+4")
        check("@for(i, a+1, b*2-3, -@{i*2})", " -6 -8 -10 -12")
        check("@for(i, b, a,-1,@i)", "543")
        check("@for(i, b, a,-1,@for(j,0,i,@i@j))", "505152535440414243303132")
        check("@{a@{a-1}+10}", "11")
        check("@{a@a}", "2")
        check("@if(0,1,0)", "0")
        check("@if(1,1,0)", "1")
        check("@if(0,1)", "")
        check("@if(1,1)", "1")
        check("@for(i,0,8,@if(i%2,+@i))", "+1+3+5+7")
        check("@{1<1}", "0")
        check("@{!1}", "0")
        check("@{!!1}", "1")
        check("@{!!1<<2}", "4")
        check("@{a<b*a1}", "1")
        check("@{a^b == 7}", "2")
        check("@{(a^b) == 7}", "1")
        check("@{b<<a == 5*4}", "1")
        check('''#include "ops/binary_op_defs.h"
        #define OP1(a, b) a+b
        OP1
        @expand_macro(OP1,1,2)
        @expand_macro(maximum, T, 1, 2)
        @expand_macro(@OP,T,1,2)''',
        '''        #define OP1(a, b) a+b
        OP1
        1+2
        std::max(T(1), T(2))
        ((1)+T(2)*(T(rcount)))''')
        expect_error(lambda: jit_precompile(vars, "@{a"))
        expect_error(lambda: jit_precompile(vars, "@for(a"))
        expect_error(lambda: jit_precompile(vars, "@for(i,l,r)"))
        expect_error(lambda: jit_precompile(vars, "@for(i,l,(@i,,,,))"))
        expect_error(lambda: jit_precompile(vars, "@for(i,0,10000,@i)"))
        expect_error(lambda: jit_precompile(vars, "@for(i,0,-1,@i)"))
        expect_error(lambda: jit_precompile(vars, "@asd"))
        expect_error(lambda: jit_precompile(vars, "@if"))
        expect_error(lambda: jit_precompile(vars, "@if(1,1,1,1)"))
        expect_error(lambda: jit_precompile(vars, "@if(1)"))
        expect_error(lambda: jit_precompile(vars, "#define OP1(a,b) a+b\n@expand_macro(OP1,1)"))

    def test_strcmp(self):
        vars = {"Tx":"float"}
        check = lambda expr, result: \
            self.assertEqual(jit_precompile(vars, expr), result)
        check("@strcmp(aaa,aaa)", "0")
        check("@strcmp(aaa,bbb)", "-1")
        check("@strcmp(ccc,bbb)", "1")
        check("@{@strcmp(aaa,aaa)}", "0")
        check("@{@strcmp(aaa,bbb)}", "-1")
        check("@{@strcmp(ccc,bbb)}", "1")

        code = \
"""@define(T_NCCL,
    @if(@strcmp(@Tx,float)==0 || @strcmp(@Tx,float32)==0, ncclFloat)
    @if(@strcmp(@Tx,int)==0 || @strcmp(@Tx,int32)==0, ncclInt)
    @if(@strcmp(@Tx,float64)==0, ncclFloat64)
    @if(@strcmp(@Tx,int64)==0, ncclInt64)
)
ncclBcast(..., @T_NCCL, ...)
"""
        assert "ncclFloat" in jit_precompile({"Tx":"float"}, code)
        assert "ncclFloat" in jit_precompile({"Tx":"float32"}, code)
        assert "ncclFloat64" in jit_precompile({"Tx":"float64"}, code)
        assert "ncclInt" in jit_precompile({"Tx":"int"}, code)
        assert "ncclInt" in jit_precompile({"Tx":"int32"}, code)
        assert "ncclInt64" in jit_precompile({"Tx":"int64"}, code)
        
    def test_mif(self):
        vars = {"Tx":"float"}
        check = lambda expr, result: \
            self.assertEqual(jit_precompile(vars, expr), result)
        check("#if aa>1\n@Tx\n#else\n@Tx@@1\n#endif", "#if aa>1\nfloat\n#else\nfloat1\n#endif")


if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2021 Jittor. All Rights Reserved. 
# Maintainers: 
#     Wenyang Zhou <576825820@qq.com>
#     Dun Liang <randonlang@gmail.com>. 
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
# This model is generated by pytorch converter.

import jittor as jt
from jittor import nn
__all__ = ['MNASNet', 'mnasnet0_5', 'mnasnet0_75', 'mnasnet1_0', 'mnasnet1_3']
_BN_MOMENTUM = (1 - 0.9997)

class _InvertedResidual(nn.Module):

    def __init__(self, in_ch, out_ch, kernel_size, stride, expansion_factor, bn_momentum=0.1):
        super(_InvertedResidual, self).__init__()
        assert (stride in [1, 2])
        assert (kernel_size in [3, 5])
        mid_ch = (in_ch * expansion_factor)
        self.apply_residual = ((in_ch == out_ch) and (stride == 1))
        self.layers = nn.Sequential(nn.Conv(in_ch, mid_ch, 1, bias=False), nn.BatchNorm(mid_ch, momentum=bn_momentum), nn.Relu(), nn.Conv(mid_ch, mid_ch, kernel_size, padding=(kernel_size // 2), stride=stride, groups=mid_ch, bias=False), nn.BatchNorm(mid_ch, momentum=bn_momentum), nn.Relu(), nn.Conv(mid_ch, out_ch, 1, bias=False), nn.BatchNorm(out_ch, momentum=bn_momentum))

    def execute(self, input):
        if self.apply_residual:
            return (self.layers(input) + input)
        else:
            return self.layers(input)

def _stack(in_ch, out_ch, kernel_size, stride, exp_factor, repeats, bn_momentum):
    assert (repeats >= 1)
    first = _InvertedResidual(in_ch, out_ch, kernel_size, stride, exp_factor, bn_momentum=bn_momentum)
    remaining = []
    for _ in range(1, repeats):
        remaining.append(_InvertedResidual(out_ch, out_ch, kernel_size, 1, exp_factor, bn_momentum=bn_momentum))
    return nn.Sequential(first, *remaining)

def _round_to_multiple_of(val, divisor, round_up_bias=0.9):
    assert (0.0 < round_up_bias < 1.0)
    new_val = max(divisor, ((int((val + (divisor / 2))) // divisor) * divisor))
    return (new_val if (new_val >= (round_up_bias * val)) else (new_val + divisor))

def _get_depths(alpha):
    depths = [24, 40, 80, 96, 192, 320]
    return [_round_to_multiple_of((depth * alpha), 8) for depth in depths]

class MNASNet(nn.Module):
    """ MNASNet model architecture. version=2.

    Args:
    
    * alpha: Depth multiplier.
    * num_classes: Number of classes. Default: 1000.
    * dropout: Dropout probability of dropout layer.
    """
    _version = 2

    def __init__(self, alpha, num_classes=1000, dropout=0.2):
        super(MNASNet, self).__init__()
        assert (alpha > 0.0)
        self.alpha = alpha
        self.num_classes = num_classes
        depths = _get_depths(alpha)
        layers = [
            nn.Conv(3, 32, 3, padding=1, stride=2, bias=False), 
            nn.BatchNorm(32, momentum=_BN_MOMENTUM), 
            nn.Relu(), 
            nn.Conv(32, 32, 3, padding=1, stride=1, groups=32, bias=False), 
            nn.BatchNorm(32, momentum=_BN_MOMENTUM), 
            nn.Relu(), 
            nn.Conv(32, 16, 1, padding=0, stride=1, bias=False), 
            nn.BatchNorm(16, momentum=_BN_MOMENTUM), 
            _stack(16, depths[0], 3, 2, 3, 3, _BN_MOMENTUM),
            _stack(depths[0], depths[1], 5, 2, 3, 3, _BN_MOMENTUM),
            _stack(depths[1], depths[2], 5, 2, 6, 3, _BN_MOMENTUM),
            _stack(depths[2], depths[3], 3, 1, 6, 2, _BN_MOMENTUM),
            _stack(depths[3], depths[4], 5, 2, 6, 4, _BN_MOMENTUM),
            _stack(depths[4], depths[5], 3, 1, 6, 1, _BN_MOMENTUM),
            nn.Conv(depths[5], 1280, 1, padding=0, stride=1, bias=False), 
            nn.BatchNorm(1280, momentum=_BN_MOMENTUM), 
            nn.Relu()
        ]
        self.layers = nn.Sequential(*layers)
        self.classifier = nn.Sequential(nn.Dropout(p=dropout), nn.Linear(1280, num_classes))

    def execute(self, x):
        x = self.layers(x)
        x = x.mean([2, 3])
        return self.classifier(x)

def mnasnet0_5(pretrained=False, **kwargs):
    model = MNASNet(0.5, **kwargs)
    if pretrained: model.load("jittorhub://mnasnet0_5.pkl")
    return model

def mnasnet0_75(pretrained=False, **kwargs):
    model = MNASNet(0.75, **kwargs)
    if pretrained: model.load("jittorhub://mnasnet0_75.pkl")
    return model

def mnasnet1_0(pretrained=False, **kwargs):
    model = MNASNet(1.0, **kwargs)
    if pretrained: model.load("jittorhub://mnasnet1_0.pkl")
    return model

def mnasnet1_3(pretrained=False, **kwargs):
    model = MNASNet(1.3, **kwargs)
    if pretrained: model.load("jittorhub://mnasnet1_3.pkl")
    return model
